@@ -1,21 +1,14 @@
-use crate::fixer::{Fixer, FixerError, SerializableFixer};
-use crate::rule::{RuleSerializeError, SerializableRule};
-use crate::transform::Transformation;
-use crate::DeserializeEnv;
 use crate::GlobalRules;
 
-pub use crate::constraints::{
-  try_deserialize_matchers, RuleWithConstraint, SerializableMetaVarMatcher,
-  SerializeConstraintsError,
-};
+use crate::fixer::Fixer;
+use crate::rule::DeserializeEnv;
+pub use crate::rule_core::{RuleConfigError, RuleCore, SerializableRuleCore};
 use ast_grep_core::language::Language;
-use ast_grep_core::meta_var::MetaVarMatchers;
-use ast_grep_core::replacer::{IndentSensitive, Replacer};
+use ast_grep_core::replacer::{Content, Replacer};
 use ast_grep_core::{NodeMatch, StrDoc};
 use schemars::JsonSchema;
 use serde::{Deserialize, Serialize};
-use serde_yaml::{with::singleton_map_recursive::deserialize, Deserializer, Error as YamlError};
-use thiserror::Error;
+use serde_yaml::{with::singleton_map_recursive::deserialize, Deserializer};
 
 use std::collections::HashMap;
 use std::ops::{Deref, DerefMut};
@@ -37,73 +30,23 @@ pub enum Severity {
 }
 
 #[derive(Serialize, Deserialize, Clone, JsonSchema)]
-pub struct SerializableRuleCore<L: Language> {
-  /// Specify the language to parse and the file extension to include in matching.
-  pub language: L,
-  /// A rule object to find matching AST nodes
-  pub rule: SerializableRule,
-  /// Additional meta variables pattern to filter matching
-  pub constraints: Option<HashMap<String, SerializableMetaVarMatcher>>,
-  /// Utility rules that can be used in `matches`
-  pub utils: Option<HashMap<String, SerializableRule>>,
-  /// A dictionary for metavariable manipulation. Dict key is the new variable name.
-  /// Dict value is a [transformation] that specifies how meta var is processed.
-  /// Warning: this is experimental option. [`https://github.com/ast-grep/ast-grep/issues/436`]
-  pub transform: Option<HashMap<String, Transformation>>,
-}
-
-impl<L: Language> SerializableRuleCore<L> {
-  fn get_deserialize_env(&self, globals: &GlobalRules<L>) -> RResult<DeserializeEnv<L>> {
-    let env = DeserializeEnv::new(self.language.clone()).with_globals(globals);
-    if let Some(utils) = &self.utils {
-      let env = env.register_local_utils(utils)?;
-      Ok(env)
-    } else {
-      Ok(env)
-    }
-  }
-
-  fn get_meta_var_matchers(&self) -> RResult<MetaVarMatchers<StrDoc<L>>> {
-    Ok(if let Some(constraints) = self.constraints.clone() {
-      try_deserialize_matchers(constraints, self.language.clone())?
-    } else {
-      MetaVarMatchers::default()
-    })
-  }
-
-  pub fn get_matcher(&self, globals: &GlobalRules<L>) -> RResult<RuleWithConstraint<L>> {
-    let env = self.get_deserialize_env(globals)?;
-    let rule = env.deserialize_rule(self.rule.clone())?;
-    let matchers = self.get_meta_var_matchers()?;
-    let transform = self.transform.clone();
-    Ok(
-      RuleWithConstraint::new(rule)
-        .with_matchers(matchers)
-        .with_utils(env.registration)
-        .with_transform(transform),
-    )
-  }
-}
-#[derive(Serialize, Deserialize, Clone, JsonSchema)]
-pub struct SerializableRuleConfigCore<L: Language> {
+pub struct SerializableRewriter {
   #[serde(flatten)]
-  pub core: SerializableRuleCore<L>,
+  pub core: SerializableRuleCore,
   /// Unique, descriptive identifier, e.g., no-unused-variable
   pub id: String,
 }
 
-pub fn into_map<L: Language>(
-  rules: Vec<SerializableRuleConfigCore<L>>,
-) -> HashMap<String, SerializableRuleCore<L>> {
-  rules.into_iter().map(|r| (r.id, r.core)).collect()
-}
-
 #[derive(Serialize, Deserialize, Clone, JsonSchema)]
 pub struct SerializableRuleConfig<L: Language> {
   #[serde(flatten)]
-  pub core: SerializableRuleCore<L>,
+  pub core: SerializableRuleCore,
   /// Unique, descriptive identifier, e.g., no-unused-variable
   pub id: String,
+  /// Specify the language to parse and the file extension to include in matching.
+  pub language: L,
+  /// Rewrite rules for `applyRewriters` transformation
+  pub rewriters: Option<Vec<SerializableRewriter>>,
   /// Main message highlighting why this rule fired. It should be single line and concise,
   /// but specific enough to be understood without additional context.
   #[serde(default)]
@@ -113,8 +56,6 @@ pub struct SerializableRuleConfig<L: Language> {
   /// One of: hint, info, warning, or error
   #[serde(default)]
   pub severity: Severity,
-  /// A pattern to auto fix the issue. It can reference metavariables appeared in rule.
-  pub fix: Option<SerializableFixer>,
   /// Glob patterns to specify that the rule only applies to matching files
   pub files: Option<Vec<String>>,
   /// Glob patterns that exclude rules from applying to files
@@ -125,28 +66,20 @@ pub struct SerializableRuleConfig<L: Language> {
   pub metadata: Option<HashMap<String, String>>,
 }
 
-type RResult<T> = std::result::Result<T, RuleConfigError>;
-
 impl<L: Language> SerializableRuleConfig<L> {
-  pub fn get_fixer<C: IndentSensitive>(&self) -> RResult<Option<Fixer<C, L>>> {
-    let transform = &self.transform;
-    if let Some(fixer) = &self.fix {
-      let env = self.get_deserialize_env(&Default::default())?;
-      let parsed = Fixer::parse(fixer, &env, transform)?;
-      Ok(Some(parsed))
-    } else {
-      Ok(None)
-    }
-  }
-
   fn get_message(&self, node_match: &NodeMatch<StrDoc<L>>) -> String {
     let bytes = self.message.generate_replacement(node_match);
     String::from_utf8(bytes).expect("replacement must be valid utf-8")
   }
+
+  pub fn get_matcher(&self, globals: &GlobalRules<L>) -> Result<RuleCore<L>, RuleConfigError> {
+    let env = DeserializeEnv::new(self.language.clone()).with_globals(globals);
+    self.core.get_matcher(env)
+  }
 }
 
 impl<L: Language> Deref for SerializableRuleConfig<L> {
-  type Target = SerializableRuleCore<L>;
+  type Target = SerializableRuleCore;
   fn deref(&self) -> &Self::Target {
     &self.core
   }
@@ -158,22 +91,9 @@ impl<L: Language> DerefMut for SerializableRuleConfig<L> {
   }
 }
 
-#[derive(Debug, Error)]
-pub enum RuleConfigError {
-  #[error("Fail to parse yaml as RuleConfig")]
-  Yaml(#[from] YamlError),
-  #[error("Rule is not configured correctly.")]
-  Rule(#[from] RuleSerializeError),
-  #[error("fix pattern is invalid.")]
-  Fixer(#[from] FixerError),
-  #[error("constraints is not configured correctly.")]
-  Constraints(#[from] SerializeConstraintsError),
-}
-
 pub struct RuleConfig<L: Language> {
   inner: SerializableRuleConfig<L>,
-  pub matcher: RuleWithConstraint<L>,
-  pub fixer: Option<Fixer<String, L>>,
+  pub matcher: RuleCore<L>,
 }
 
 impl<L: Language> RuleConfig<L> {
@@ -181,13 +101,22 @@ impl<L: Language> RuleConfig<L> {
     inner: SerializableRuleConfig<L>,
     globals: &GlobalRules<L>,
   ) -> Result<Self, RuleConfigError> {
-    let matcher = inner.get_matcher(globals)?;
-    let fixer = inner.get_fixer()?;
-    Ok(Self {
-      inner,
-      matcher,
-      fixer,
-    })
+    let mut matcher = inner.get_matcher(globals)?;
+    let Some(ser) = inner.rewriters.clone() else {
+      return Ok(Self { inner, matcher });
+    };
+    let mut rewriters = HashMap::new();
+    for val in ser {
+      // NB should inherit env from matcher to inherit utils
+      // TODO: optimize duplicate env creation/util registraion
+      let env = DeserializeEnv::new(inner.language.clone()).with_globals(globals);
+      let env = inner.get_deserialize_env(env)?;
+      let env = val.core.get_deserialize_env(env)?;
+      let rewriter = val.core.get_matcher_from_env(&env)?;
+      rewriters.insert(val.id, rewriter);
+    }
+    matcher.add_rewriters(rewriters)?;
+    Ok(Self { inner, matcher })
   }
 
   pub fn deserialize<'de>(
@@ -204,6 +133,15 @@ impl<L: Language> RuleConfig<L> {
   pub fn get_message(&self, node: &NodeMatch<StrDoc<L>>) -> String {
     self.inner.get_message(node)
   }
+  pub fn get_fixer<C: Content>(&self) -> Result<Option<Fixer<L>>, RuleConfigError> {
+    if let Some(fix) = &self.fix {
+      let env = self.matcher.get_env(self.language.clone());
+      let parsed = Fixer::parse(fix, &env, &self.transform)?;
+      Ok(Some(parsed))
+    } else {
+      Ok(None)
+    }
+  }
 }
 impl<L: Language> Deref for RuleConfig<L> {
   type Target = SerializableRuleConfig<L>;
@@ -216,23 +154,25 @@ impl<L: Language> Deref for RuleConfig<L> {
 mod test {
   use super::*;
   use crate::from_str;
+  use crate::rule::SerializableRule;
   use crate::test::TypeScript;
 
   fn ts_rule_config(rule: SerializableRule) -> SerializableRuleConfig<TypeScript> {
     let core = SerializableRuleCore {
-      language: TypeScript::Tsx,
       rule,
       constraints: None,
       transform: None,
       utils: None,
+      fix: None,
     };
     SerializableRuleConfig {
       core,
       id: "".into(),
+      language: TypeScript::Tsx,
+      rewriters: None,
       message: "".into(),
       note: None,
       severity: Severity::Hint,
-      fix: None,
       files: None,
       ignores: None,
       url: None,
@@ -381,4 +321,126 @@ test-rule:
     let grep = TypeScript::Tsx.ast_grep("some()");
     assert!(grep.root().find(&matcher).is_none());
   }
+  #[test]
+  fn test_get_fixer() {
+    let globals = GlobalRules::default();
+    let mut config = get_matches_config();
+    config.fix = Some(from_str("string!!").unwrap());
+    let rule = RuleConfig::try_from(config, &globals).unwrap();
+    let fixer = rule.get_fixer::<String>().unwrap().unwrap();
+    let grep = TypeScript::Tsx.ast_grep("some(123)");
+    let nm = grep.root().find(&rule.matcher).unwrap();
+    let replacement = fixer.generate_replacement(&nm);
+    assert_eq!(String::from_utf8_lossy(&replacement), "string!!");
+  }
+
+  #[test]
+  fn test_add_rewriters() {
+    let rule: SerializableRuleConfig<TypeScript> = from_str(
+      r"
+id: test
+rule: {pattern: 'a = $A'}
+language: Tsx
+transform:
+  B:
+    applyRewriters:
+      rewriters: [re]
+      source: $A
+rewriters:
+- id: re
+  rule: {kind: number}
+  fix: yjsnp
+    ",
+    )
+    .expect("should parse");
+    let rule = RuleConfig::try_from(rule, &Default::default()).expect("work");
+    let grep = TypeScript::Tsx.ast_grep("a = 123");
+    let nm = grep.root().find(&rule.matcher).unwrap();
+    let b = nm.get_env().get_transformed("B").expect("should have");
+    assert_eq!(String::from_utf8_lossy(b), "yjsnp");
+  }
+
+  #[test]
+  fn test_rewriters_access_utils() {
+    let rule: SerializableRuleConfig<TypeScript> = from_str(
+      r"
+id: test
+rule: {pattern: 'a = $A'}
+language: Tsx
+utils:
+  num: { kind: number }
+transform:
+  B:
+    applyRewriters:
+      rewriters: [re]
+      source: $A
+rewriters:
+- id: re
+  rule: {matches: num, pattern: $NOT}
+  fix: yjsnp
+    ",
+    )
+    .expect("should parse");
+    let rule = RuleConfig::try_from(rule, &Default::default()).expect("work");
+    let grep = TypeScript::Tsx.ast_grep("a = 456");
+    let nm = grep.root().find(&rule.matcher).unwrap();
+    let b = nm.get_env().get_transformed("B").expect("should have");
+    assert!(nm.get_env().get_match("NOT").is_none());
+    assert_eq!(String::from_utf8_lossy(b), "yjsnp");
+  }
+
+  #[test]
+  fn test_rewriter_utils_should_not_pollute_registration() {
+    let rule: SerializableRuleConfig<TypeScript> = from_str(
+      r"
+id: test
+rule: {matches: num}
+language: Tsx
+transform:
+  B:
+    applyRewriters:
+      rewriters: [re]
+      source: $A
+rewriters:
+- id: re
+  rule: {matches: num}
+  fix: yjsnp
+  utils:
+    num: { kind: number }
+    ",
+    )
+    .expect("should parse");
+    let rule = RuleConfig::try_from(rule, &Default::default()).expect("work");
+    let grep = TypeScript::Tsx.ast_grep("a = 456");
+    let nm = grep.root().find(&rule.matcher);
+    assert!(nm.is_none());
+  }
+
+  #[test]
+  fn test_utils_in_rewriter_should_work() {
+    let rule: SerializableRuleConfig<TypeScript> = from_str(
+      r"
+id: test
+rule: {pattern: 'a = $A'}
+language: Tsx
+transform:
+  B:
+    applyRewriters:
+      rewriters: [re]
+      source: $A
+rewriters:
+- id: re
+  rule: {matches: num}
+  fix: yjsnp
+  utils:
+    num: { kind: number }
+    ",
+    )
+    .expect("should parse");
+    let rule = RuleConfig::try_from(rule, &Default::default()).expect("work");
+    let grep = TypeScript::Tsx.ast_grep("a = 114514");
+    let nm = grep.root().find(&rule.matcher).unwrap();
+    let b = nm.get_env().get_transformed("B").expect("should have");
+    assert_eq!(String::from_utf8_lossy(b), "yjsnp");
+  }
 }
